#!/bin/sh

srcdir=$(dirname $0)
test -z "$srcdir" && srcdir=.

# Generate the documentation about compiling Window Maker
$srcdir/script/generate-txt-from-texi.sh "$srcdir/doc/build/Compilation.texi" \
					 -o "$srcdir/INSTALL-WMAKER"
$srcdir/script/generate-txt-from-texi.sh "$srcdir/doc/build/Translations.texi" \
					 -o "$srcdir/README.i18n"

# Change date of the files to the past so they will be regenerated by 'make'
touch -d '2000-01-01' $srcdir/INSTALL-WMAKER $srcdir/README.i18n

# Generate the configure script from the 'configure.ac'
autoreconf -vfi -I m4 $srcdir
